@@ -1,5 +1,5 @@
 /* Copyright (C) 1995, 1996, 1998, 2000, 2001, 2004, 2006,
- *   2008-2015 Free Software Foundation, Inc.
+ *   2008-2016 Free Software Foundation, Inc.
  * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -51,6 +51,9 @@
 /* {Strings}
  */
 
+SCM_SYMBOL (sym_UTF_8, "UTF-8");
+SCM_SYMBOL (sym_ISO_8859_1, "ISO-8859-1");
+SCM_SYMBOL (sym_error, "error");
 
 /* Stringbufs 
  *
@@ -1611,11 +1614,18 @@ scm_from_locale_string (const char *str)
   return scm_from_locale_stringn (str, -1);
 }
 
+scm_t_string_failed_conversion_handler
+scm_i_default_string_failed_conversion_handler (void)
+{
+  return scm_i_string_failed_conversion_handler
+    (scm_i_default_port_conversion_strategy ());
+}
+
 SCM
 scm_from_locale_stringn (const char *str, size_t len)
 {
   return scm_from_stringn (str, len, locale_charset (),
-                           scm_i_default_port_conversion_handler ());
+                           scm_i_default_string_failed_conversion_handler ());
 }
 
 SCM
@@ -1757,17 +1767,18 @@ scm_from_port_string (const char *str, SCM port)
 SCM
 scm_from_port_stringn (const char *str, size_t len, SCM port)
 {
-  scm_t_port *pt = SCM_PTAB_ENTRY (port);
-  scm_t_port_internal *pti = SCM_PORT_GET_INTERNAL (port);
+  scm_t_port *pt = SCM_PORT (port);
 
-  if (pti->encoding_mode == SCM_PORT_ENCODING_MODE_LATIN1)
+  if (scm_is_eq (pt->encoding, sym_ISO_8859_1))
     return scm_from_latin1_stringn (str, len);
-  else if (pti->encoding_mode == SCM_PORT_ENCODING_MODE_UTF8
-           && (pt->ilseq_handler == SCM_FAILED_CONVERSION_ERROR
+  else if (scm_is_eq (pt->encoding, sym_UTF_8)
+           && (scm_is_eq (pt->conversion_strategy, sym_error)
                || (u8_check ((uint8_t *) str, len) == NULL)))
     return scm_from_utf8_stringn (str, len);
   else
-    return scm_from_stringn (str, len, pt->encoding, pt->ilseq_handler);
+    return scm_from_stringn (str, len, scm_i_symbol_chars (pt->encoding),
+                             scm_i_string_failed_conversion_handler
+                             (scm_port_conversion_strategy (port)));
 }
 
 /* Create a new scheme string from the C string STR.  The memory of
@@ -1938,7 +1949,7 @@ scm_to_locale_stringn (SCM str, size_t *lenp)
 {
   return scm_to_stringn (str, lenp,
                          locale_charset (),
-                         scm_i_default_port_conversion_handler ());
+                         scm_i_default_string_failed_conversion_handler ());
 }
 
 char *
@@ -2054,6 +2065,38 @@ u32_u8_length_in_bytes (const scm_t_uint32 *str, size_t len)
   return ret;
 }
 
+static size_t
+utf8_length (SCM str)
+{
+  if (scm_i_is_narrow_string (str))
+    return latin1_u8_strlen ((scm_t_uint8 *) scm_i_string_chars (str),
+                             scm_i_string_length (str));
+  else
+    return u32_u8_length_in_bytes
+      ((scm_t_uint32 *) scm_i_string_wide_chars (str),
+       scm_i_string_length (str));
+}
+
+size_t
+scm_c_string_utf8_length (SCM string)
+#define FUNC_NAME "scm_c_string_utf8_length"
+{
+  SCM_VALIDATE_STRING (1, string);
+  return utf8_length (string);
+}
+#undef FUNC_NAME
+
+SCM_DEFINE (scm_string_utf8_length, "string-utf8-length", 1, 0, 0,
+	    (SCM string),
+	    "Returns the number of bytes in the UTF-8 representation of "
+            "@var{string}.")
+#define FUNC_NAME s_scm_string_utf8_length
+{
+  SCM_VALIDATE_STRING (1, string);
+  return scm_from_size_t (utf8_length (string));
+}
+#undef FUNC_NAME
+
 char *
 scm_to_utf8_stringn (SCM str, size_t *lenp)
 #define FUNC_NAME "scm_to_utf8_stringn"
@@ -2164,16 +2207,17 @@ scm_to_port_string (SCM str, SCM port)
 char *
 scm_to_port_stringn (SCM str, size_t *lenp, SCM port)
 {
-  scm_t_port *pt = SCM_PTAB_ENTRY (port);
-  scm_t_port_internal *pti = SCM_PORT_GET_INTERNAL (port);
+  scm_t_port *pt = SCM_PORT (port);
 
-  if (pti->encoding_mode == SCM_PORT_ENCODING_MODE_LATIN1
-      && pt->ilseq_handler == SCM_FAILED_CONVERSION_ERROR)
+  if (scm_is_eq (pt->encoding, sym_ISO_8859_1)
+      && scm_is_eq (pt->conversion_strategy, sym_error))
     return scm_to_latin1_stringn (str, lenp);
-  else if (pti->encoding_mode == SCM_PORT_ENCODING_MODE_UTF8)
+  else if (scm_is_eq (pt->encoding, sym_UTF_8))
     return scm_to_utf8_stringn (str, lenp);
   else
-    return scm_to_stringn (str, lenp, pt->encoding, pt->ilseq_handler);
+    return scm_to_stringn (str, lenp, scm_i_symbol_chars (pt->encoding),
+                           scm_i_string_failed_conversion_handler
+                           (scm_port_conversion_strategy (port)));
 }
 
 /* Return a malloc(3)-allocated buffer containing the contents of STR encoded
