@@ -1,6 +1,6 @@
 ;;; Repl commands
 
-;; Copyright (C) 2001, 2009, 2010, 2011 Free Software Foundation, Inc.
+;; Copyright (C) 2001, 2009, 2010, 2011, 2012, 2013 Free Software Foundation, Inc.
 
 ;; This library is free software; you can redistribute it and/or
 ;; modify it under the terms of the GNU Lesser General Public
@@ -25,11 +25,11 @@
   #:use-module (system base compile)
   #:use-module (system repl common)
   #:use-module (system repl debug)
-  #:use-module (system vm objcode)
+  #:use-module (system vm disassembler)
+  #:use-module (system vm loader)
   #:use-module (system vm program)
   #:use-module (system vm trap-state)
   #:use-module (system vm vm)
-  #:use-module ((system vm frame) #:select (frame-return-values))
   #:autoload (system base language) (lookup-language language-reader)
   #:autoload (system vm trace) (call-with-trace)
   #:use-module (ice-9 format)
@@ -40,6 +40,7 @@
   #:use-module (ice-9 control)
   #:use-module ((ice-9 pretty-print) #:select ((pretty-print . pp)))
   #:use-module ((system vm inspect) #:select ((inspect . %inspect)))
+  #:use-module (rnrs bytevectors)
   #:use-module (statprof)
   #:export (meta-command define-meta-command))
 
@@ -342,22 +343,29 @@ Find bindings/modules/packages."
 (define-meta-command (describe repl (form))
   "describe OBJ
 Show description/documentation."
-  (display (object-documentation (repl-eval repl (repl-parse repl form))))
+  (display
+    (object-documentation
+      (let ((input (repl-parse repl form)))
+        (if (symbol? input)
+            (module-ref (current-module) input)
+            (repl-eval repl input)))))
   (newline))
 
 (define-meta-command (option repl . args)
-  "option [KEY VALUE]
+  "option [NAME] [EXP]
 List/show/set options."
   (pmatch args
     (()
      (for-each (lambda (spec)
 		 (format #t "  ~A~24t~A\n" (car spec) (cadr spec)))
 	       (repl-options repl)))
-    ((,key)
-     (display (repl-option-ref repl key))
+    ((,name)
+     (display (repl-option-ref repl name))
      (newline))
-    ((,key ,val)
-     (repl-option-set! repl key val))))
+    ((,name ,exp)
+     ;; Would be nice to evaluate in the current language, but the REPL
+     ;; option parser doesn't permit that, currently.
+     (repl-option-set! repl name (eval exp (current-module))))))
 
 (define-meta-command (quit repl)
   "quit
@@ -441,6 +449,7 @@ Change languages."
         (cur (repl-language repl)))
     (format #t "Happy hacking with ~a!  To switch back, type `,L ~a'.\n"
             (language-title lang) (language-name cur))
+    (current-language lang)
     (set! (repl-language repl) lang)))
 
 
@@ -448,11 +457,15 @@ Change languages."
 ;;; Compile commands
 ;;;
 
+(define (load-image x)
+  (let ((thunk (load-thunk-from-memory x)))
+    (find-mapped-elf-image (program-code thunk))))
+
 (define-meta-command (compile repl (form))
   "compile EXP
 Generate compiled code."
   (let ((x (repl-compile repl (repl-parse repl form))))
-    (cond ((objcode? x) (guile:disassemble x))
+    (cond ((bytevector? x) (disassemble-image (load-image x)))
           (else (repl-print repl x)))))
 
 (define-meta-command (compile-file repl file . opts)
@@ -474,22 +487,24 @@ Run the optimizer on a piece of code and print the result."
     (run-hook before-print-hook x)
     (pp x)))
 
-(define (guile:disassemble x)
-  ((@ (language assembly disassemble) disassemble) x))
-
 (define-meta-command (disassemble repl (form))
   "disassemble EXP
 Disassemble a compiled procedure."
   (let ((obj (repl-eval repl (repl-parse repl form))))
-    (if (or (program? obj) (objcode? obj))
-        (guile:disassemble obj)
-        (format #t "Argument to ,disassemble not a procedure or objcode: ~a~%"
-                obj))))
+    (cond
+     ((program? obj)
+      (disassemble-program obj))
+     ((bytevector? obj)
+      (disassemble-image (load-image obj)))
+     (else
+      (format #t
+              "Argument to ,disassemble not a procedure or a bytevector: ~a~%"
+              obj)))))
 
 (define-meta-command (disassemble-file repl file)
   "disassemble-file FILE
 Disassemble a file."
-  (guile:disassemble (load-objcode (->string file))))
+  (disassemble-file (->string file)))
 
 
 ;;;
@@ -549,8 +564,6 @@ Trace execution."
                        (identifier-syntax (debug-frames debug)))
                       (#,(datum->syntax #'repl 'message)
                        (identifier-syntax (debug-error-message debug)))
-                      (#,(datum->syntax #'repl 'for-trap?)
-                       (identifier-syntax (debug-for-trap? debug)))
                       (#,(datum->syntax #'repl 'index)
                        (identifier-syntax
                         (id (debug-index debug))
@@ -572,8 +585,7 @@ If COUNT is negative, the last COUNT frames will be shown."
   (print-frames frames
                 #:count count
                 #:width width
-                #:full? full?
-                #:for-trap? for-trap?))
+                #:full? full?))
 
 (define-stack-command (up repl #:optional (count 1))
   "up [COUNT]
@@ -590,12 +602,10 @@ An argument says how many frames up to go."
       (format #t "Already at outermost frame.\n"))
      (else
       (set! index (1- (vector-length frames)))
-      (print-frame cur #:index index
-                   #:next-source? (and (zero? index) for-trap?)))))
+      (print-frame cur #:index index))))
    (else
     (set! index (+ count index))
-    (print-frame cur #:index index
-                 #:next-source? (and (zero? index) for-trap?)))))
+    (print-frame cur #:index index))))
 
 (define-stack-command (down repl #:optional (count 1))
   "down [COUNT]
@@ -612,11 +622,10 @@ An argument says how many frames down to go."
       (format #t "Already at innermost frame.\n"))
      (else
       (set! index 0)
-      (print-frame cur #:index index #:next-source? for-trap?))))
+      (print-frame cur #:index index))))
    (else
     (set! index (- index count))
-    (print-frame cur #:index index
-                 #:next-source? (and (zero? index) for-trap?)))))
+    (print-frame cur #:index index))))
 
 (define-stack-command (frame repl #:optional idx)
   "frame [IDX]
@@ -631,17 +640,10 @@ With an argument, select a frame by index, then show it."
       (format #t "Invalid argument to `frame': expected a non-negative integer for IDX.~%"))
      ((< idx (vector-length frames))
       (set! index idx)
-      (print-frame cur #:index index
-                   #:next-source? (and (zero? index) for-trap?)))
+      (print-frame cur #:index index))
      (else
       (format #t "No such frame.~%"))))
-   (else (print-frame cur #:index index
-                      #:next-source? (and (zero? index) for-trap?)))))
-
-(define-stack-command (procedure repl)
-  "procedure
-Print the procedure for the selected frame."
-  (repl-print repl (frame-procedure cur)))
+   (else (print-frame cur #:index index))))
 
 (define-stack-command (locals repl #:key (width (terminal-width)))
   "locals
@@ -680,8 +682,8 @@ Note that the given source location must be inside a procedure."
       (format #t "Trap ~a: ~a.~%" idx (trap-name idx)))))
 
 (define (repl-pop-continuation-resumer repl msg)
-  ;; Capture the dynamic environment with this prompt thing. The
-  ;; result is a procedure that takes a frame.
+  ;; Capture the dynamic environment with this prompt thing. The result
+  ;; is a procedure that takes a frame and number of values returned.
   (% (call-with-values
          (lambda ()
            (abort
@@ -689,20 +691,20 @@ Note that the given source location must be inside a procedure."
               ;; Call frame->stack-vector before reinstating the
               ;; continuation, so that we catch the %stacks fluid at
               ;; the time of capture.
-              (lambda (frame)
+              (lambda (frame . values)
                 (k frame
                    (frame->stack-vector
-                    (frame-previous frame)))))))
-       (lambda (from stack)
+                    (frame-previous frame))
+                   values)))))
+       (lambda (from stack values)
          (format #t "~a~%" msg)
-         (let ((vals (frame-return-values from)))
-           (if (null? vals)
-               (format #t "No return values.~%")
-               (begin
-                 (format #t "Return values:~%")
-                 (for-each (lambda (x) (repl-print repl x)) vals))))
+         (if (null? values)
+             (format #t "No return values.~%")
+             (begin
+               (format #t "Return values:~%")
+               (for-each (lambda (x) (repl-print repl x)) values)))
          ((module-ref (resolve-interface '(system repl repl)) 'start-repl)
-          #:debug (make-debug stack 0 msg #t))))))
+          #:debug (make-debug stack 0 msg))))))
 
 (define-stack-command (finish repl)
   "finish
@@ -726,7 +728,7 @@ Resume execution, breaking when the current frame finishes."
                        (k (frame->stack-vector frame)))))))
        (format #t "~a~%" msg)
        ((module-ref (resolve-interface '(system repl repl)) 'start-repl)
-        #:debug (make-debug stack 0 msg #t)))))
+        #:debug (make-debug stack 0 msg)))))
 
 (define-stack-command (step repl)
   "step
