@@ -24,7 +24,8 @@
   #:use-module (system base language)
   #:use-module (system base message)
   #:use-module (system vm program)
-  #:autoload (language tree-il optimize) (optimize!)
+  #:use-module (system vm loader)
+  #:autoload (language tree-il optimize) (optimize)
   #:use-module (ice-9 control)
   #:use-module (ice-9 history)
   #:export (<repl> make-repl repl-language repl-options
@@ -175,7 +176,7 @@ See <http://www.gnu.org/licenses/lgpl.html>, for more details.")
 (define (repl-compile repl form)
   (let ((from (repl-language repl))
         (opts (repl-compile-options repl)))
-    (compile form #:from from #:to 'objcode #:opts opts
+    (compile form #:from from #:to 'bytecode #:opts opts
              #:env (current-module))))
 
 (define (repl-expand repl form)
@@ -188,10 +189,10 @@ See <http://www.gnu.org/licenses/lgpl.html>, for more details.")
 (define (repl-optimize repl form)
   (let ((from (repl-language repl))
         (opts (repl-compile-options repl)))
-    (decompile (optimize! (compile form #:from from #:to 'tree-il #:opts opts
-                                   #:env (current-module))
-                          (current-module)
-                          opts)
+    (decompile (optimize (compile form #:from from #:to 'tree-il #:opts opts
+                                  #:env (current-module))
+                         (current-module)
+                         opts)
                #:from 'tree-il #:to from)))
 
 (define (repl-parse repl form)
@@ -204,7 +205,7 @@ See <http://www.gnu.org/licenses/lgpl.html>, for more details.")
              (or (null? (language-compilers (repl-language repl)))
                  (repl-option-ref repl 'interp)))
         (lambda () (eval form (current-module)))
-        (make-program (repl-compile repl form)))))
+        (load-thunk-from-memory (repl-compile repl form)))))
 
 (define (repl-eval repl form)
   (let ((thunk (repl-prepare-eval-thunk repl form)))
